﻿/*==Description=========================================================================

QPX()
Tests how fast functions and/or subroutines run and diplays results in a window.

Author:		Learning one
Contact:	http://learningone.ahk4.net/Learning one contact.png
Thanks to:	SKAN
AHK forum:	http://www.autohotkey.com/board/topic/91860-object-qpx-test-how-fast-the-code-runs/
Version:	1.02



License.
You can freely use and distribute this product if you give a credit to the author.

Credits.
Thanks to SKAN for his QPX function; www.autohotkey.com/forum/viewtopic.php?t=52083

See also.
QPX() by SKAN;  http://www.autohotkey.com/board/topic/48063-qpx-delay-based-on-queryperformancecounter/
AutoHotkey Benchmarks by Uberi; http://www.autohotkey.com/board/topic/62971-autohotkey-benchmarks/
Minimal Benchmark by R3gX; http://www.autohotkey.com/board/topic/91860-object-qpx-test-how-fast-the-code-runs/#entry581412

Documentation.
QPX() tests how fast functions and/or subroutines run and diplays results in a window.
When you are in doubt which coding approach/style will do specific job fastest, this test will give you the answer.
Call Add() method to add function(s) and/or subroutine(s) to the test collection, and after that,
call the Test() method to test how fast they run and display results in a window. Left edit = Summary. Right edit = Details.
See comments in the code for more details.
User doesn't have to call anything to construct QPX object because it initializes itself automatically before the auto-execute section of the script.
User is not supposed to derive any object from QPX base/class object.

Basic usage concept.
QPX.Add("MyFunc")							; adds MyFunc function to the test collection
QPX.Add("AnotherFunc", "param1", "param2")	; adds AnotherFunc function to the test collection and passes 2 parameters to that function
QPX.Add("MySub")							; adds MySub subroutine to the test collection
QPX.Add("AnotherSub")						; adds AnotherSub subroutine to the test collection
QPX.Test()									; starts test and displays results in a window. Left edit = Summary. Right edit = Details.
return





Example script 1 - Tests how fast will each function do its job - reverse a comma delimited list
	;===Auto-execute===
	Loop, 100		; makes a comma delimited list for testing; "1,2,3, ...etc."
		List .= A_Index ","
	List := RTrim(List, ",")

	QPX.Add("ReverseList1", List)	; adds ReverseList1 function to a test collection and passes List as parameter to it.
	QPX.Add("ReverseList2", List)	; adds ReverseList2 function to a test collection and passes List as parameter to it.
	QPX.Test()						; starts test and displays results in a window. Left edit = Summary. Right edit = Details.
	return

	;===Hotkeys===
	F1::QPX.Test(50)				; starts test again, but this time with 50 iterations (instead of 10 - default)
	Esc::ExitApp

	;===Functions===
	#Include QPX.ahk	; by Learning one
	ReverseList1(List, Delimiter=",") {
		StringSplit, p, List, % Delimiter
		Loop, % p0
		{
			CurItemNum := (A_Index = 1) ? p0 : p0 - A_Index + 1
			OutList .= p%CurItemNum% Delimiter
		}
		return RTrim(OutList, Delimiter)
	}
	ReverseList2(List, Delimiter=",") {	; from the Sort command documentation in AHK kelp
		Sort, List, F ReverseList2_Helper D%Delimiter%  ; Reverses the list so that it contains 4,3,2,1
		return List
	}
	ReverseList2_Helper(a1, a2, offset) { ; a helper function
		return offset  ; Offset is positive if a2 came after a1 in the original list; negative otherwise.
	}






Example script 2 - Tests how fast will each function do its job - remove duplicate lines from a string
	;===Auto-execute===
	; Let's build a test string which will have duplicate lines
	Loop, 100	; adds unique lines to the string
		String .= A_Index "`n"
	Loop, 50	; makes random duplicate lines in a string
	{
		Random, ran, 1, 100
		String .= ran "`n"
	}
	
	; Let's see which function will do its job fastest - remove duplicate lines from a string. 
	QPX.Add("RemoveDuplicates1", String)	; adds RemoveDuplicates1 function to a test collection and passes String as parameter to it.
	QPX.Add("RemoveDuplicates2", String)	; adds RemoveDuplicates2 function to a test collection and passes String as parameter to it.
	QPX.Add("RemoveDuplicates3", String)	; adds RemoveDuplicates3 function to a test collection and passes String as parameter to it.
	QPX.Test()								; starts test and displays results in a window. Left edit = Summary. Right edit = Details.
	return

	;===Hotkeys===	; just to demonstrate some additional options and functionalities (not necessary)
	F1::QPX.Test(100)							; starts test again, but this time with 100 iterations (instead of 10 - default)
	F2::QPX.Remove("RemoveDuplicates1")			; removes RemoveDuplicates1 function from a test collection
	F3::QPX.Remove()							; removes all items (functions and/or subroutines) from the test collection
	F4::QPX.Add("RemoveDuplicates1", String)	; adds RemoveDuplicates1 function to a test collection (again) and passes String as parameter to it.
	Esc::ExitApp

	;===Functions===
	#Include QPX.ahk	; by Learning one
	RemoveDuplicates1(String, Delimiter="`n") {		; by rbrtryn - http://www.autohotkey.com/board/topic/91682-how-to-remove-duplicated-lines-with-regex/#entry578368
		Loop, parse, String, % Delimiter
			if not RegExMatch(out, "\b" A_LoopField "\b") {
				out .= out ? "`n" : ""
				out .= A_LoopField
			}
			return out
	}
	RemoveDuplicates2(String, Delimiter="`n") {		; by sinkfaze - http://www.autohotkey.com/board/topic/91682-how-to-remove-duplicated-lines-with-regex/#entry578382
		oUniques := []
		Loop, parse, String, % Delimiter
			if	!oUniques[A_LoopField]
				oUniques[A_LoopField] := 1, NewString .= Delimiter A_LoopField
		return	LTrim(NewString, Delimiter)
	}
	RemoveDuplicates3(String, Delimiter="`n") {		; by polyethene (slightly modified) - http://www.autohotkey.com/board/topic/91682-how-to-remove-duplicated-lines-with-regex/#entry578384
		Loop, Parse, String, % Delimiter
			If not InStr(list, new := RegExReplace(A_LoopField, "[\d\-:. ]+(.*?)[/ ].*", "$1") . Delimiter)
				Str .= A_LoopField . Delimiter, list .= new
			return Str
	}

*/


Class QPX {	; Tests how fast functions and/or subroutines run and diplays results in a window. By Learning one. Uses QPX function by SKAN - thank you!
	static FuncRef := [], FuncParams := [], SubNames := []	; collections of: function references, function parameters, subroutine names
	static AutoInit := QPX.__New()	; AutoInit is evaluated before the auto-execute section (Static Initializer), and it constructs QPX object.
	__New() {
		w := 400, h := 250
		Gui, New, +HwndhGui
		Gui %hGui%:Font, , Courier New
		Gui %hGui%:Add, Text, x22 y22 w22 h22 Hide hwndhgDefocuser
		Gui %hGui%:Add, Edit, % "x2 y2 w" (w-4)/2 " h" h-6 " hwndhgEdit1"
		Gui %hGui%:Add, Edit,  % "x" w/2 " y2 w" (w-4)/2 " h" h-6 " hwndhgEdit2"
		Gui %hGui%:Show, % "w" w " h" h " Hide", QPC test
		this.hGui := hGui, this.hgDefocuser := hgDefocuser, this.hgEdit1 := hgEdit1, this.hgEdit2 := hgEdit2	; store
	}
	Add(params*) {	; Adds a function or a subroutine to the test collection.
	/*
	This method automatically detects are you adding a function or a subroutine to the test collection. If you are adding a :
	a)	subroutine, 1. parameter is a subroutine name. Example: QPX.Add("MySubroutine")
	b)	function, 1. parameter is a function name, and other parameters (optional) are that function's parameters. If parameter has expressions, all of them are evaluated in the moment when they are added to the collection (not in the moment when Test method is called). Example1: QPX.Add("MyFunction") Example2: QPX.Add("AnotherFunction", "param1", "param2", "param3")
	If you have a function and a subroutine with the same name in your code, and you are adding it to the test collection, function will be added (not subroutine).
	*/		
		if (IsFunc(params.1) > 0) {	; a function
			this.FuncRef.Insert(Func(params.1))	; convert function name to function reference and store it in this.FuncRef
			NewParams := []
			For k,v in params
			{
				if (k != 1)	; 1. param is the function name, other are function params
					NewParams.Insert(v)
			}
			this.FuncParams.Insert(NewParams)	; store function parameters
			;this.FuncRef.1.(this.FuncParams.1*)	; call 1 function in collection with its parameters
			;MsgBox % this.FuncRef.2.(this.FuncParams.2*)	; call 2 function in collection with its parameters and display return value in MsgBox
		}
		else if (IsLabel(params.1) > 0) {	; a subroutine
			this.SubNames.Insert(params.1)	; store subroutine name
			;Gosub, % this.SubNames.1	; execute 1. subroutine in collection
		}
	}
	Remove(FuncOrSubName="") {	;  If FuncOrSubName is blank, it removes all functions and subroutines from the test collection, otherwise, it removes specified func or sub.
		if (FuncOrSubName= "") {	; remove all functions and subs from the collection - overwrite with blank obj
			this.FuncRef := [], this.FuncParams := [], this.SubNames := []	; WRONG would be: this.FuncRef := "", this.FuncParams := "", this.SubNames := ""
		}
		else {	; remove specified func or sub from collection
			if (IsFunc(FuncOrSubName) > 0) {	; function
				For k,v in this.FuncRef			; v = function reference
				{
					if (v.Name = FuncOrSubName) {	; we want to remove this item
						ToRemove := k
						break
					}
				}
				if (ToRemove != "")
					this.FuncRef.Remove(ToRemove), this.FuncParams.Remove(ToRemove)	; remove from both func collections!
			}
			else if (IsLabel(FuncOrSubName) > 0) {	; subroutine
				For k,v in this.SubNames			; v = subroutine name
				{
					if (v = FuncOrSubName) {	; we want to remove this item
						ToRemove := k
						break
					}
				}
				if (ToRemove != "")
					this.SubNames.Remove(ToRemove)	; remove from sub collection
			}
		}
	}
	Test(NumberOfIterations=10, Sleep=10) {	; Tests functions and/or subroutines in test collection and displays results in a window. Left edit = Summary. Right edit = Details.
		oElapsedTimesF := []	; collection of elapsed times for all functions. Each item in collection is a sub-object (simple array) which consist of all elapsed times for that function
		oElapsedTimesS := []	; collection of elapsed times for all subroutines. Each item in collection is a sub-object (simple array) which consist of all elapsed times for that subroutine
		; oElapsedTimesF.3 = simple array of elapsed times for 3. function
		; oElapsedTimesF.3.4 = get number which represents 4. elapsed time in 3. function
		
		hGui := this.hGui
		
		if (this.FuncRef.MaxIndex() = "" and this.SubNames.MaxIndex() = "") {	; there's nothing in test collection
			Gui %hGui%:+OwnDialogs
			MsgBox,, QPC test, There is nothing to test. Please use Add method to add functions or subroutines you want to test.
			return
		}
		For k,v in this.FuncRef	; create item (sub-object) for each function in collection which will be populated with elapsed times
			oElapsedTimesF.Insert([])
		For k,v in this.SubNames	; create item (sub-object) for each subroutine in collection which will be populated with elapsed times
			oElapsedTimesS.Insert([])	
		FastestAve := 1000000000000	; here we'll store the fastest average run time of all items in test collection
		
		GuiControl, % this.hGui ":+ReadOnly" , % this.hgEdit1, ; Summary
		GuiControl, % this.hGui ":+ReadOnly" , % this.hgEdit2, ; Details
		Gui %hGui%:Show
		
		Loop, % NumberOfIterations			; test for specified number of iterations
		{
			For k,v in this.FuncRef			; test each function. v = function reference
			{
				this.QPX(1)					; Start - Initialise Counter
				v.(this.FuncParams[k]*)		; call a function (by reference) with its parameters. To see return value each time: MsgBox % v.(this.FuncParams[k]*)
				ElapsedTime := this.QPX(0)	; Stop, get time consumed and reset internal vars
				
				oElapsedTimesF[k].Insert(ElapsedTime)	; insert elapsed time for this function
				Sleep, % Sleep				; Good: more objective and accurate results + anti high CPU load. Bad: test lasts longer...
			}
			For k,v in this.SubNames		; test each subroutine. v = subroutine name
			{
				this.QPX(1)					; Start - Initialise Counter
				Gosub, % v					; execute subroutine in collection
				ElapsedTime := this.QPX(0)	; Stop, get time consumed and reset internal vars
				
				oElapsedTimesS[k].Insert(ElapsedTime)	; insert elapsed time for this subroutine
				Sleep, % Sleep				; Good: more objective and accurate results + anti high CPU load. Bad: test lasts longer...
			}
			
			WinSetTitle, % "ahk_id " this.hGui,, % "QPC test - " A_Index "/" NumberOfIterations " iterations"
		}
		
		For k,v in oElapsedTimesF	; process function test results. v = reference to simple array of elapsed times for current function
		{
			Min := 1000000000000, Max := 0, Tot := 0
			For k2,v2 in v			; v2 = current elapsed time
			{
				if (v2 < Min)	; new minimum
					Min := v2
				if (v2 > Max)	; new maximum
					Max := v2
				Tot += v2		; add time to a total time elapsed
			}
			Ave	:= Tot/NumberOfIterations	; get average
			if (Ave < FastestAve)	; new fastest average run time of all items in test collection
				FastestAve := Ave, FastestAveItemName := this.FuncRef[k].Name
			
			;=== Get percentage for minimum and maximum relative to average run time, format, alignment ===
			PercMin := this.FormatPerc(Min/ave*100), PercAve := this.FormatPerc(100), PercMax := this.FormatPerc(Max/ave*100)
			
			Summary .= this.FuncRef[k].Name "`n" Ave "`n"
			Details .= this.FuncRef[k].Name "`nMin: " Min PercMin "`nAve: " Ave PercAve "`nMax: " Max PercMax "`nTot: " Tot "`n`n"
		}
		For k,v in oElapsedTimesS	; process subroutine test results. v = reference to simple array of elapsed times for current subroutine
		{
			Min := 1000000000000, Max := 0, Tot := 0
			For k2,v2 in v			; v2 = current elapsed time
			{
				if (v2 < Min)	; new minimum
					Min := v2
				if (v2 > Max)	; new maximum
					Max := v2
				Tot += v2		; add time to a total time elapsed
			}
			Ave	:= Tot/NumberOfIterations	; get average
			if (Ave < FastestAve)	; new fastest average run time of all items in test collection
				FastestAve := Ave, FastestAveItemName := this.SubNames[k]
			
			;=== Get percentage for minimum and maximum relative to average run time, format, alignment ===
			PercMin := this.FormatPerc(Min/ave*100), PercAve := this.FormatPerc(100), PercMax := this.FormatPerc(Max/ave*100)
			
			Summary .= this.SubNames[k] "`n" Ave "`n"
			Details .= this.SubNames[k] "`nMin: " Min PercMin "`nAve: " Ave PercAve "`nMax: " Max PercMax "`nTot: " Tot "`n`n"
		}
		Summary := RTrim(Summary, "`n"), Details := RTrim(Details, "`n")
		Summary := this.RefineSummary(Summary, FastestAve, FastestAveItemName)
		
		GuiControl, % this.hGui ":" , % this.hgEdit1, % Summary
		GuiControl, % this.hGui ":" , % this.hgEdit2, % Details
		GuiControl, % this.hGui ":-ReadOnly" , % this.hgEdit1,	; Summary
		GuiControl, % this.hGui ":-ReadOnly" , % this.hgEdit2,	; Details
		GuiControl, % this.hGui ":Focus", % this.hgDefocuser	; Defocus edit control + anti "all text in edit is selected" measure. Looks better :)
		Gui %hGui%:Show, , % "QPC test results - " NumberOfIterations " iterations"	; show results
	}
	
	;=== Private methods ===
	RefineSummary(Summary, FastestAve, FastestAveItemName) {	; Get percentage for each average run time relative to fastest average run time for all items in test collection, mark fastest item, indentation, and write that to summary string
		FastestAve := Trim(FastestAve)	; although there is nothing to trim, for some strange reason this has to be done
		;ToolTip % "x" FastestAve "x"
		Indent := "   "
		Loop, parse, Summary, `n, `r
		{
			if (Mod(A_Index, 2) = 1) {	; odd number (neparni broj) = Item name. Exa: "TestFunc1"
				if (A_LoopField = FastestAveItemName)	; mark this item as fastest
					NewSummary .= A_LoopField " <<`n"
				else
					NewSummary .= A_LoopField "`n"
			}
			else						; even number (parni broj) = Item's average run time. Exa: "0.001161"
				NewSummary .= Indent A_LoopField this.FormatPerc(A_LoopField/FastestAve*100) "`n"
		}
		return RTrim(NewSummary, "`n")
	}
	FormatPerc(PercentWithoutSign) {	; round number, prepend some spaces for alignment and add percent sign. Exa: in: "83.134231" out: "   83%"
		Percent := Round(PercentWithoutSign) "`%"
		WantedStrLen := 6, PerStrLen := StrLen(Percent)
		if (PerStrLen < WantedStrLen) {
			Loop % WantedStrLen - PerStrLen
				SpaceIndent .= A_Space
		}
		else
			SpaceIndent := A_Space
		return SpaceIndent Percent
	}
	QPX( N=0 ) {       ;  Wrapper for  QueryPerformanceCounter()by SKAN  | CD: 06/Dec/2009
		Static F,A,Q,P,X  ;  www.autohotkey.com/forum/viewtopic.php?t=52083 | LM: 10/Dec/2009
		If ( N && !P )
		Return  DllCall("QueryPerformanceFrequency",Int64P,F) + (X:=A:=0)
			+ DllCall("QueryPerformanceCounter",Int64P,P)
		DllCall("QueryPerformanceCounter",Int64P,Q), A:=A+Q-P, P:=Q, X:=X+1
		Return ( N && X=N ) ? (X:=X-1)<<64 : ( N=0 && (R:=A/X/F) ) ? ( R + (A:=P:=X:=0) ) : 1
	}
}